package org.btcollider.cnc.keysrv.keytree;

import java.util.BitSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Random;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import org.btcollider.cnc.CnC;
import org.btcollider.cnc.dto.KeyRange;
import org.btcollider.cnc.keysrv.KeyServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class KeyTreeServer implements KeyServer {
  final Logger log = LoggerFactory.getLogger(KeyTreeServer.class);

  private KeyTree root;
  private int index;
  private int depth;

  public KeyTreeServer(int index, int depth, long maxWorkSpan, long pruneKey) {
    assert Math.pow(2, CnC.MAX_BITS) <= Long.MAX_VALUE;
    assert index < CnC.MAX_BITS && depth <= CnC.MAX_BITS;

    this.depth = depth;
    this.index = index;
    this.root = generateKeyTree(depth);

    if (pruneKey > 0)
      prune(pruneKey);

    if (maxWorkSpan > 0) {
      log.info("Starting KeyTree concierge with timeout {}s",
          TimeUnit.MILLISECONDS.toSeconds(maxWorkSpan));
      ScheduledExecutorService es = Executors.newSingleThreadScheduledExecutor();
      es.scheduleWithFixedDelay(new KTConcierge(root, maxWorkSpan), 0, 30, TimeUnit.SECONDS);
    } else {
      log.info("KeyTree concierge disabled");
    }
  }

  @Override
  public KeyRange getRange() {
    BitSet keyStart = new BitSet(CnC.MAX_BITS);

    if (!recCollectKey(root, keyStart, index))
      return null;

    BitSet keyEnd = new BitSet(CnC.MAX_BITS);
    keyEnd.or(keyStart); // set keyEnd to keyStart
    keyEnd.flip(0, index + 1 - depth); // reverse last bits

    return new KeyRange(keyStart, keyEnd);
  }

  public void setSearched(Long from, Long to) {
    assert from >= 0 && to >= 0 && from <= to;

    long step = (long) Math.pow(2, ((index + 1) - depth));

    for (long i = from; i < to; i += step) {
      BitSet searchedKey = BitSet.valueOf(new long[] {i});
      markSearched(searchedKey);
    }
  }

  @Override
  public void setSearched(KeyRange keyRange) {
    markSearched(BitSet.valueOf(new long[] {keyRange.getStart()}));
  }

  @Override
  public void setInWork(KeyRange keyRange) {
    if (keyRange == null)
      return;

    markInWork(BitSet.valueOf(new long[] {keyRange.getStart()}));
  }

  public String printDot() {
    StringBuilder buffer = new StringBuilder();
    buffer.append("strict graph { \n");

    Queue<KeyTree> frontier = new LinkedList<>();
    frontier.add(root);

    while (!frontier.isEmpty()) {
      KeyTree node = frontier.poll();
      buffer.append(dotNodeName(node) + ";\n");

      if (node.getParent() != null) {
        buffer.append(dotNodeName(node.getParent()) + " -- " + dotNodeName(node) + ";\n");
      }

      if (node.getLeft() != null)
        frontier.add(node.getLeft());
      if (node.getRight() != null)
        frontier.add(node.getRight());
    }

    buffer.append("\n}");
    return buffer.toString();
  }

  private String dotNodeName(KeyTree node) {
    String nodeName = "\"" + node.hashCode() + "_" + node.getValue();

    if (node.isSearched())
      nodeName += "_s";
    if (node.inWork())
      nodeName += "_w";

    nodeName += "\"";

    return nodeName;
  }

  /**
   * Prunes every branch strictly smaller than pruneKey
   *
   * @param prune The smallest valid key in the keyspace Must be at least as large as the smallest
   *        key in the tree Must be at most as large as the largest key in the tree
   */
  private void prune(Long pruneKey) {
    BitSet pkBS = BitSet.valueOf(new long[] {pruneKey});
    int curIdx = pkBS.length() - 1;
    assert (curIdx == this.index); // at least the smallest, at most the largest key in the tree
                                   // -> pos of highest bits is equal

    recPrune(pkBS, this.root, curIdx);
  }

  private void recPrune(BitSet pruneKey, KeyTree node, int curIdx) {
    if (!node.getValue() && pruneKey.get(curIdx)) { // node[ix] = 0 and pruneK[ix] = 1 -> path
                                                    // smaller
      node.setSearched(true);
      return;
    } else if (node.getValue() && !pruneKey.get(curIdx)) { // node[ix] = 1 and pruneK[ix] = 0 ->
                                                           // path
      return;
    }

    if (node.isLeaf()) { // end reached
      assert (curIdx == 0);
      return;
    }

    recPrune(pruneKey, node.getLeft(), curIdx - 1);
    recPrune(pruneKey, node.getRight(), curIdx - 1);
  }

  private void markInWork(BitSet key) {
    KeyTree keyTreePointer = root;

    int curIdx = key.length() - 2; // ignore root bit
    while (curIdx >= 0) {
      if (key.get(curIdx))
        keyTreePointer = keyTreePointer.getRight();
      else
        keyTreePointer = keyTreePointer.getLeft();

      if (keyTreePointer.inWork()) {
        break; // path already marked as inWork
      }

      if (keyTreePointer.isLeaf()) { // end of tree reached
        keyTreePointer.setInWork();
        recMarkInWorkParents(keyTreePointer);
        break;
      }

      curIdx--;
    }
  }

  private void recMarkInWorkParents(KeyTree child) {
    if (child.getParent() == null)
      return;

    KeyTree parent = child.getParent();

    if (!parent.getLeft().inWork() || !parent.getRight().inWork()) {
      return;
    }

    // Mark parent inWork too, check next level
    parent.setInWork(Long.max(parent.getLeft().inWorkSince(), parent.getRight().inWorkSince()));

    recMarkInWorkParents(parent);
  }

  private void markSearched(BitSet key) {
    KeyTree keyTreePointer = root;

    int curIdx = key.length() - 2; // ignore root bit
    while (curIdx >= 0) {
      if (key.get(curIdx))
        keyTreePointer = keyTreePointer.getRight();
      else
        keyTreePointer = keyTreePointer.getLeft();

      if (keyTreePointer.isSearched())
        break; // path already marked as searched

      if (keyTreePointer.isLeaf()) { // end of tree reached
        keyTreePointer.setSearched(true);
        recMarkSearchedParents(keyTreePointer);
        break;
      }

      curIdx--;
    }
  }

  private void recMarkSearchedParents(KeyTree child) {
    if (child.getParent() == null)
      return;

    KeyTree parent = child.getParent();

    if (!parent.getLeft().isSearched() || !parent.getRight().isSearched()) {
      return;
    }

    // Mark parent searched too, check next level
    parent.setSearched(true);
    recMarkSearchedParents(parent);
  }

  private KeyTree generateKeyTree(int depth) {
    assert depth > 0;

    KeyTree genRoot = new KeyTree(true, null);
    recGenKeyTree(genRoot, depth - 1);

    return genRoot;
  }

  private void recGenKeyTree(KeyTree kt, int depth) {
    assert depth >= 0;

    if (depth == 0)
      return;

    KeyTree left = new KeyTree(false, kt);
    KeyTree right = new KeyTree(true, kt);

    kt.setLeft(left);
    kt.setRight(right);

    recGenKeyTree(left, depth - 1);
    recGenKeyTree(right, depth - 1);
  }

  private boolean recCollectKey(KeyTree kt, BitSet keyStart, int curIndex) {
    assert curIndex > 0;

    keyStart.set(curIndex, kt.getValue());

    KeyTree left = kt.getLeft();
    KeyTree right = kt.getRight();

    // Finish criteria
    if (left.isLeaf() || right.isLeaf()) {
      assert left.isLeaf() && right.isLeaf(); // fully balanced binary tree by construction

      if (!right.isFree() && !left.isFree())
        return false;

      // No choice finish
      if (!right.isFree())
        keyStart.set(curIndex - 1, left.getValue());
      else if (!left.isFree())
        keyStart.set(curIndex - 1, right.getValue());
      else if (left.isFree() && right.isFree()) { // Randomized finish
        Random r = new Random();
        boolean leftFirst = r.nextBoolean();
        if (leftFirst)
          keyStart.set(curIndex - 1, left.getValue());
        else
          keyStart.set(curIndex - 1, right.getValue());
      }

      return true;
    }

    // No choice depth-first traversal
    if (!left.isFree())
      return recCollectKey(right, keyStart, curIndex - 1);
    else if (!right.isFree())
      return recCollectKey(left, keyStart, curIndex - 1);
    else { // Randomized depth-first traversal
      Random r = new Random();
      boolean leftFirst = r.nextBoolean();
      if (leftFirst)
        return recCollectKey(left, keyStart, curIndex - 1);
      else
        return recCollectKey(right, keyStart, curIndex - 1);
    }
  }

  protected KeyTree getKeyTree() {
    return this.root;
  }
}
